# -*- coding: utf-8 -*-
"""Common enum and common setup functions."""

import logging
import argparse
from enum import Enum
from collections import namedtuple


Purpose = Enum('Purpose', 'Feature BugFix Chore HotFix Unknown')
ReviewNumPlace = Enum('ReviewNumPlace', 'head tail none')
Mode = Enum('Mode', 'md html')
MergeInfo = namedtuple('MergeInfo', (
    'commit',
    'author',
    'body',
    'review_num',
    'review_url',
    'purpose')
)

DESCRIPTIOON = \
    'grou stands for Git Release note OUtput command.\n' \
    'Currently grou output the release note info from git merge log.\n'\
    'Othre git log info like "commitizen" style are not supported for now.'


def setup_logger(enable_debug):
    """Log level setting from command line args."""
    loglevel = logging.ERROR
    if enable_debug:
        loglevel = logging.DEBUG
    logging.basicConfig(format='%(asctime)s [%(levelname)s]'
                               ' %(message)s', level=loglevel)


def setup_commandline_args():
    """Setup argparse and Taple of regurn command line args."""
    parser = argparse.ArgumentParser(description=DESCRIPTIOON)
    parser.add_argument('-f', '--from',
                        type=str,
                        dest='f',
                        help='Git commitish as the "from" point'
                             ' for seach merge logs.'
                             ' This point is not included to logs.',
                        required=False)
    parser.add_argument('-t', '--to',
                        type=str,
                        dest='t',
                        help='Git commitish as the "to" point'
                             ' for seach merge logs. Default is HEAD',
                        default='HEAD',
                        required=False)
    parser.add_argument('-s', '--style',
                        type=str,
                        help='Output style. Default is md (markdown).',
                        choices=['md', 'html'],
                        default='md',
                        required=False)
    parser.add_argument('-b', '--branch',
                        type=str,
                        help='Git branch name for searching merge logs.'
                             ' Default is master.',
                        default='master',
                        required=False)
    parser.add_argument('--debug',
                        action='store_true',
                        default=False,
                        help='Show debug log if this flag is set'
                        ' (default: False)')
    parser.add_argument('--use-commit-body',
                        action='store_true',
                        default=False,
                        help='Use commit log body to'
                             ' lines of the release list.'
                             ' If not set, commit log subject will be used.'
                             ' (default: False)')
    parser.add_argument('--no-author',
                        action='store_true',
                        default=False,
                        help='Stop to add commit author name for each line.'
                             ' (default: False)')
    parser.add_argument('--no-link',
                        action='store_true',
                        default=False,
                        help='Stop to add link for each review number.'
                             ' (default: False)')
    parser.add_argument('--no-hash',
                        action='store_true',
                        default=False,
                        help='Stop to add commit hash for each line.'
                             ' (default: False)')
    parser.add_argument('--show-review-num',
                        type=str,
                        help='The place where the review number should be put.'
                             ' Default is head.',
                        choices=['head', 'tail', 'none'],
                        default='head',
                        required=False)
    parser.add_argument('--remove-regix',
                        type=str,
                        help='Regix to remove specific texts from each line.',
                        required=False)
    parser.add_argument('--remove-regix-i',
                        type=str,
                        help='Regix to remove specific texts from each line.',
                        required=False)
    return parser.parse_args()



# -*- coding: utf-8 -*-
"""Formatter classes and their factory class."""
from builtins import super
from common import Mode, ReviewNumPlace
from abc import ABCMeta, abstractmethod


class OutputFomatterFactory():
    """Create suitable formatter for style."""

    @classmethod
    def create(cls, style_name):
        """Create suitable formatter for style."""
        if style_name == Mode.md.name:
            return MarkdownOutputFormatter()
        elif style_name == Mode.html.name:
            return HtmlOutputFormatter()
        else:
            raise ValueError('Unexpected style: ' + style_name)


class AbstractOutputFormatter():
    """Abstract class for formatter."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def create_section_title(self, purpose_name):
        """Raw string for section title."""
        pass

    @abstractmethod
    def create_list_start(self):
        """Raw string for the start of list."""
        pass

    @abstractmethod
    def create_list_end(self):
        """Raw string for the end of list."""
        pass

    @abstractmethod
    def create_list(self, info, options):
        """Raw string for a list."""
        pass

    def put_review_num(self, show_review_num_name, texts, review_num_texts):
        """Insert or append review number and link."""
        if show_review_num_name == ReviewNumPlace.head.name:
            texts.insert(1, ''.join(review_num_texts))
        elif show_review_num_name == ReviewNumPlace.tail.name:
            texts.extend(review_num_texts)

        return texts


class MarkdownOutputFormatter(AbstractOutputFormatter):
    """Formatter class for Markdown style."""

    def create_section_title(self, purpose_name):
        """See super class description."""
        return '# ' + purpose_name

    def create_list_start(self):
        """See super class description."""
        return ''

    def create_list_end(self):
        """See super class description."""
        return '\n'

    def create_list(self, info, options):
        """See super class description."""
        texts = []
        review_num_texts = []

        texts.append('-')

        if options.no_hash is False:
            texts.append(' `{0.commit}`'.format(info))

        texts.append(u' {0.body}'.format(info))

        if options.no_author is False:
            texts.append(' by {0.author}'.format(info))

        # create review number and link
        review_num_texts.append(' [PR{0.review_num}]'.format(info))
        if options.no_link is False:
            review_num_texts.append('({0.review_url})'.format(info))
        # put review number at the suitable index of texts.
        texts = super().put_review_num(options.show_review_num,
                                       texts,
                                       review_num_texts)
        return ''.join(texts)


class HtmlOutputFormatter(AbstractOutputFormatter):
    """Formatter class for HTML style."""

    def create_section_title(self, purpose_name):
        """See super class description."""
        return '<h1>' + purpose_name + '</h1>'

    def create_list_start(self):
        """See super class description."""
        return '<ul>'

    def create_list_end(self):
        """See super class description."""
        return '</ul>'

    def create_list(self, info, options):
        """See super class description."""
        texts = []
        review_num_texts = []

        texts.append('<li>')
        if options.no_hash is False:
            texts.append(' <code>{0.commit}</code>'.format(info))
        texts.append(' {0.body}'.format(info))
        if options.no_author is False:
            texts.append(' by {0.author}'.format(info))

        # create review number and link
        if options.no_link is False:
            review_num_texts.append('<a href="{0.review_url}">'.format(info))
        review_num_texts.append('PR{0.review_num}'.format(info))
        if options.no_link is False:
            review_num_texts.append('</a>'.format(info))
        texts.append('</li>')

        # put review number at the suitable index of texts.
        texts = super().put_review_num(options.show_review_num,
                                       texts,
                                       review_num_texts)
        return ''.join(texts)



# -*- coding: utf-8 -*-
"""Converter for raw string log to Merge Info."""

import logging
import re
from git_info_converter import GithubInfoConverter
from common import Purpose, MergeInfo


def convert_to_info(base_url, line, options):
    """Convert 1 line log to MergeInfo."""
    logging.debug('rawline = ' + str(line))
    # create dictionary for each merge log.
    splitted_log = line.decode('utf-8').split(':')
    # Get PR number
    if len(splitted_log[2]) == 0 or len(splitted_log[3]) == 0 \
            or re.search('#[0-9]*', splitted_log[3]) is None:
        logging.debug('Ignore unexpected merge log: ' + str(line))
        return None, None
    review_num = GithubInfoConverter.pickup_review_num(splitted_log[3])
    url = GithubInfoConverter.convert_to_review_url(base_url, review_num)
    # Detect PR type
    purpose = Purpose.Unknown
    body = splitted_log[2]
    if bool(re.match('Feature/', body, re.I)):
        purpose = Purpose.Feature
    elif bool(re.match('Bug[fix]*/', body, re.I)):
        purpose = Purpose.BugFix
    elif bool(re.match('Chore/', body, re.I)):
        purpose = Purpose.Chore
    elif bool(re.match('HotFix/', body, re.I)):
        purpose = Purpose.HotFix

    # Format body
    if options.use_commit_body is False:
        body = splitted_log[3]
    elif purpose != Purpose.Unknown:
        body = body.split('/', 1)[1]

    # remove text by using option
    if options.remove_regix:
        logging.debug('before = ' + body)
        body = re.sub(options.remove_regix, '', body).strip()
        logging.debug('after  = ' + body)
    if options.remove_regix_i:
        logging.debug('before = ' + body)
        body = re.sub(options.remove_regix_i, '', body, flags=re.I).strip()
        logging.debug('after  = ' + body)

    info = MergeInfo(commit=str(splitted_log[0]),
                     author=splitted_log[1],
                     body=body.title(),
                     review_url=url,
                     review_num=review_num,
                     purpose=purpose)
    logging.debug(info)
    return (purpose, info)



"""Setup.py for pip distribution."""

from setuptools import setup
from grou.common import DESCRIPTIOON

requires = ['enum34',
            'future']

setup(
    name='grou',
    version='0.1.5',
    description=DESCRIPTIOON,
    url='https://github.com/gki/grou',
    author='Gen Takeda',
    author_email='gki.penguin@gmail.com',
    license='MIT',
    keywords='git merges log formatted pretty release note output',
    packages=[
        "grou",
    ],
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=requires,
    entry_points={
        'console_scripts':
            ['grou=grou.command_line:main']
    },
)



# -*- coding: utf-8 -*-
"""Main script for grou."""

# TODO Test
# TODO pip release

import logging
import git_info_converter
import merge_info_converter
from merge_info_formatter import OutputFomatterFactory
from common import Purpose, setup_commandline_args, setup_logger


def main():
    """Main function of this tool."""
    args = setup_commandline_args()
    setup_logger(args.debug)
    base_url = git_info_converter.create_base_url()
    lines = git_info_converter.create_git_merge_logs(args.branch,
                                                     args.f,
                                                     args.t)
    release_dict = {purpose: [] for purpose in Purpose}
    for line in lines:
        purpose, info = merge_info_converter.convert_to_info(base_url,
                                                             line,
                                                             args)
        if purpose is None or info is None:
            continue
        release_dict[purpose].append(info)

    # Convert to target style
    formatter = OutputFomatterFactory.create(args.style)
    exists_output = False
    for purpose in release_dict:
        if len(release_dict[purpose]) == 0:
            continue
        print(formatter.create_section_title(purpose.name))
        print(formatter.create_list_start())
        for info in release_dict[purpose]:
            print(formatter.create_list(info, args))
        print(formatter.create_list_end())
        exists_output = True

    if exists_output is False:
        logging.error('Could not output any formatted log. '
                      'Please check git merge log subject or body.\n')

        exit(1)



# -*- coding: utf-8 -*-
"""Converter for the response strings from git commands."""

import subprocess
import sys
import logging
import re


def create_git_merge_logs(branch_name, commitish_from, commitish_to):
    """Create the log text list from git log command."""
    cmd = 'git log --first-parent' \
          + ' --merges --pretty=format:%h:%an:%b:%s'
    cmd = cmd.split(' ')

    # Set from-to tag or commit id
    if commitish_from is not None:
        cmd.append(commitish_from + '..' + commitish_to)
    logging.debug(cmd)
    merges = subprocess.check_output(cmd)

    if len(merges) == 0:
        logging.error('There is no merge logs.')
        sys.exit(1)

    return merges.splitlines()


def create_base_url():
    """Create the base web url from git remote info."""
    # get organization and repo name from git config.
    cmd = 'git config remote.origin.url'.split(' ')
    remote_url = subprocess.check_output(cmd)
    if len(remote_url) == 0:
        logging.error('No remote.origin.url setting for git. '
                      'Please set origin.\n'
                      'e.g.) '
                      'git remote add origin '
                      'git@github.com:yourorg/yourrepo.git')
        sys.exit(1)

    return remote_url.decode('utf-8') \
                     .replace('\n', '') \
                     .replace(':', '/') \
                     .replace('git@', 'https://') \
                     .replace('.git', '')


class GithubInfoConverter():
    """Converter class for Github info.

    This kind of classes and super class should be added
    when support other Git repo service like GitLab.
    """

    @classmethod
    def pickup_review_num(cls, line):
        """Pickup a review number from 1 line log."""
        return re.search('#[0-9]*', line).group(0)

    @classmethod
    def convert_to_review_url(cls, base_url, review_num):
        """Create a web url from review number."""
        return base_url + '/pull/' + review_num.strip('#')
